Attribute VB_Name = "modGraphics"
' Technical info
'
'   Bit Depths
'   ----------
'   Mode 0: 4 colors, 2 bit per pixel(bpp)
'   Mode 1: 8 colors, 3 bpp (you may see this in some tile editors but the snes doesn't display
'       3bpp graphics. They have to be expanded to the 4bpp format before being used.)
'   Mode 2: 16 colors, 4 bpp
'   Mode 3: 128/256 colors, 8bpp (only in Mode 7)

'   Tile format
'   -----------
'
'   The tile data itself will consist of these bytes:
'
'   .db $00, $81, $81, $42, $C3, $24, $E7, $18, $E7, $18, $C3, $24, $81, $42, $00, $81
'
'
'   Each row of pixels is represented by two bytes. Take the first row ($00, $81). It consists
'   of the colors 2,0,0,0,0,0,0,2. In binary form:
'
'   0 0 0 0 0 0 0 0
'   1 0 0 0 0 0 0 1
'
'   Each of the bits in byte 0 are equivelent to bit 0 in it's corresponding pixel. Take the
'   next row ($81, $42):
'
'   10000001        $81,
'   01000010        $42,
'   12000021    colors
'
'   You may hear people refer to bitplanes. Those mean sets of bytes corresponding to bit ?'s
'   of all the pixels, where ? is from 0 to the number of bits per pixel minus 1. Thus, in this
'   case, bitplane 0 would refer to the odd bytes in the tile data($00,$81,$C3,etc.).
'
'   Palette format
'   --------------
'
'   Colors for the snes are 15 bits each, and stored in bgr format.
'
'       %.bbbbbgg gggrrrrr
'        |||||||| |||+++++-red
'        ||||||++-+++------green
'        |+++++------------blue
'        +-----------------UNUSED
'
'   The color full green like this would be stored as:
'
'       %00000011, %11100000    or  $03, $E0
'
'

Option Explicit




' this function coverts "numOfTiles" amount of tiles from "src"
' starting at "addr." all data is written to "dest."
'   4bpp -> 8bpp (non-interleaved)
Public Function convertFrom4BPP(ByRef src() As Byte, _
                ByVal srcaddr As Long, ByVal numOfTiles As Integer, _
                ByRef dest() As Byte) As Boolean
    
    Dim lenOfArray As Long
    Dim i As Long, ixTile As Integer, cBitMask As Byte
    Dim ix_1 As Long, ix_2 As Long, ix_3 As Long, ix_4 As Long
    
    On Error GoTo err
    
    ' resize array if too small
    lenOfArray = (numOfTiles * 64)
    If (UBound(dest) < lenOfArray) Then
        ReDim dest(lenOfArray)
    End If
    
    ' convert data
    For i = 0 To (numOfTiles - 1)
        For ixTile = 0 To 7
            cBitMask = 128
            ix_1 = (i * 32) + (ixTile * 2): ix_2 = (i * 32) + (ixTile * 2) + 1
            ix_3 = (i * 32) + (ixTile * 2) + 16: ix_4 = (i * 32) + (ixTile * 2) + 17

            dest(i * 64 + ixTile * 8 + 0) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 1) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 2) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 3) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 4) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 5) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 6) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
            dest(i * 64 + ixTile * 8 + 7) = ((src(srcaddr + ix_1) And cBitMask) / cBitMask) Or (((src(srcaddr + ix_2) And cBitMask) / cBitMask) * 2) Or (((src(srcaddr + ix_3) And cBitMask) / cBitMask) * 4) Or (((src(srcaddr + ix_4) And cBitMask) / cBitMask) * 8): cBitMask = RShift(cBitMask, 1)
        Next ixTile
    Next i

    convertFrom4BPP = True
    Exit Function
err:
    convertFrom4BPP = False
End Function

' this function coverts "numOfTiles" amount of tiles from "src"
' starting at "addr." all data is written to "dest."
'   2bpp -> 8bpp (non-interleaved)
Public Function convertFrom2BPP(ByRef src() As Byte, _
                ByVal srcaddr As Long, ByVal numOfTiles As Integer, _
                ByRef dest() As Byte) As Boolean

    Dim lenOfArray As Long
    Dim byte0 As Byte, byte1 As Byte
    Dim t As Integer, r As Integer, c As Integer
    
    On Error GoTo err

    ' resize array if too small
    lenOfArray = (numOfTiles * 64)
    If (UBound(dest) < lenOfArray) Then
        ReDim dest(lenOfArray)
    End If

    ' convert data
    For t = 0 To (numOfTiles - 1)
        For r = 0 To 7
            byte0 = src(srcaddr + (t * 16) + r)
            byte1 = src(srcaddr + (t * 16) + r + 8)
            dest((t * 64) + (r * 8) + 0) = (byte0 And &H1) + ((byte1 And &H1) * &H2)
            dest((t * 64) + (r * 8) + 1) = ((byte0 And &H2) / &H2) + (byte1 And &H2)
            dest((t * 64) + (r * 8) + 2) = ((byte0 And &H4) / &H4) + ((byte1 And &H4) / &H2)
            dest((t * 64) + (r * 8) + 3) = ((byte0 And &H8) / &H8) + ((byte1 And &H8) / &H4)
            dest((t * 64) + (r * 8) + 4) = ((byte0 And &H10) / &H10) + ((byte1 And &H10) / &H8)
            dest((t * 64) + (r * 8) + 5) = ((byte0 And &H20) / &H20) + ((byte1 And &H20) / &H10)
            dest((t * 64) + (r * 8) + 6) = ((byte0 And &H40) / &H40) + ((byte1 And &H40) / &H20)
            dest((t * 64) + (r * 8) + 7) = ((byte0 And &H80) / &H80) + ((byte1 And &H80) / &H40)
        Next r
    Next t

    convertFrom2BPP = True
    Exit Function

err:
    convertFrom2BPP = False

End Function






Private Function LShift(ByVal pnValue As Long, ByVal pnShift As Long)
    LShift = CByte((pnValue * (2 ^ pnShift)) And 255)
End Function

Private Function RShift(ByVal pnValue As Long, ByVal pnShift As Long)
    RShift = CByte((pnValue \ (2 ^ pnShift)) And 255)
End Function






Public Function convertTo4BPP()
    ' good shit
End Function

Public Function convertTo2BPP()
    ' good shit
End Function























' Input: hex value from the nes palette
' Output: the RGB value of the nes palette
Public Function Hex2Color(ByVal HexVal As Byte, ByVal monochrome As Boolean, ByVal emphasis As Byte) As Long

    Dim emp_r As Double, emp_g As Double, emp_b As Double

    
    If (monochrome) Then
        ' account for greyscale mode
        HexVal = (HexVal And &H30)
        emp_r = 1: emp_g = 1: emp_b = 1
    Else
        ' emphasis factors, %000 to %111. r,g,b
        ' measurement by Quietust
        ' {1.00, 1.00, 1.00}
        ' {1.00, 0.80, 0.81}
        ' {0.78, 0.94, 0.66}
        ' {0.79, 0.77, 0.63}
        ' {0.82, 0.83, 1.12}
        ' {0.81, 0.71, 0.87}
        ' {0.68, 0.79, 0.79}
        ' {0.70, 0.70, 0.70}
        Select Case (emphasis And 7)
            Case 0: emp_r = 1: emp_g = 1: emp_b = 1
            Case 1: emp_r = 1: emp_g = 0.8: emp_b = 0.81
            Case 2: emp_r = 0.78: emp_g = 0.94: emp_b = 0.66
            Case 3: emp_r = 0.79: emp_g = 0.77: emp_b = 0.63
            Case 4: emp_r = 0.82: emp_g = 0.83: emp_b = 1.12
            Case 5: emp_r = 0.81: emp_g = 0.71: emp_b = 0.87
            Case 6: emp_r = 0.68: emp_g = 0.79: emp_b = 0.79
            Case 7: emp_r = 0.7: emp_g = 0.7: emp_b = 0.7
        End Select
    End If

    ' get an approximation of the rgb value of PPU palette entry
    Select Case HexVal
        Case &H0
            Hex2Color = RGB(&H80 * emp_r, &H80 * emp_g, &H80 * emp_b)
        Case &H1
            Hex2Color = RGB(&H0 * emp_r, &H3D * emp_g, &HA6 * emp_b)
        Case &H2
            Hex2Color = RGB(&H0 * emp_r, &H12 * emp_g, &HB0 * emp_b)
        Case &H3
            Hex2Color = RGB(&H44 * emp_r, &H0 * emp_g, &H96 * emp_b)
        Case &H4
            Hex2Color = RGB(&HA1 * emp_r, &H0 * emp_g, &H5E * emp_b)
        Case &H5
            Hex2Color = RGB(&HC7 * emp_r, &H0 * emp_g, &H28 * emp_b)
        Case &H6
            Hex2Color = RGB(&HBA * emp_r, &H6 * emp_g, &H0 * emp_b)
        Case &H7
            Hex2Color = RGB(&H8C * emp_r, &H17 * emp_g, &H0 * emp_b)
        Case &H8
            Hex2Color = RGB(&H5C * emp_r, &H2F * emp_g, &H0 * emp_b)
        Case &H9
            Hex2Color = RGB(&H10 * emp_r, &H45 * emp_g, &H0 * emp_b)
        Case &HA
            Hex2Color = RGB(&H5 * emp_r, &H4A * emp_g, &H0 * emp_b)
        Case &HB
            Hex2Color = RGB(&H0 * emp_r, &H47 * emp_g, &H2E * emp_b)
        Case &HC
            Hex2Color = RGB(&H0 * emp_r, &H41 * emp_g, &H66 * emp_b)
        Case &HD
            Hex2Color = RGB(&H0 * emp_r, &H0 * emp_g, &H0 * emp_b)
        Case &HE
            Hex2Color = RGB(&H5 * emp_r, &H5 * emp_g, &H5 * emp_b)
        Case &HF
            Hex2Color = RGB(&H5 * emp_r, &H5 * emp_g, &H5 * emp_b)
        Case &H10
            Hex2Color = RGB(&HC7 * emp_r, &HC7 * emp_g, &HC7 * emp_b)
        Case &H11
            Hex2Color = RGB(&H0 * emp_r, &H77 * emp_g, &HFF * emp_b)
        Case &H12
            Hex2Color = RGB(&H21 * emp_r, &H55 * emp_g, &HFF * emp_b)
        Case &H13
            Hex2Color = RGB(&H82 * emp_r, &H37 * emp_g, &HFA * emp_b)
        Case &H14
            Hex2Color = RGB(&HEB * emp_r, &H2F * emp_g, &HB5 * emp_b)
        Case &H15
            Hex2Color = RGB(&HFF * emp_r, &H29 * emp_g, &H50 * emp_b)
        Case &H16
            Hex2Color = RGB(&HFF * emp_r, &H22 * emp_g, &H0 * emp_b)
        Case &H17
            Hex2Color = RGB(&HD6 * emp_r, &H32 * emp_g, &H0 * emp_b)
        Case &H18
            Hex2Color = RGB(&HC4 * emp_r, &H62 * emp_g, &H0 * emp_b)
        Case &H19
            Hex2Color = RGB(&H35 * emp_r, &H80 * emp_g, &H0 * emp_b)
        Case &H1A
            Hex2Color = RGB(&H5 * emp_r, &H8F * emp_g, &H0 * emp_b)
        Case &H1B
            Hex2Color = RGB(&H0 * emp_r, &H8A * emp_g, &H55 * emp_b)
        Case &H1C
            Hex2Color = RGB(&H0 * emp_r, &H99 * emp_g, &HCC * emp_b)
        Case &H1D
            Hex2Color = RGB(&H21 * emp_r, &H21 * emp_g, &H21 * emp_b)
        Case &H1E
            Hex2Color = RGB(&H9 * emp_r, &H9 * emp_g, &H9 * emp_b)
        Case &H1F
            Hex2Color = RGB(&H9 * emp_r, &H9 * emp_g, &H9 * emp_b)
        Case &H20
            Hex2Color = RGB(&HFF * emp_r, &HFF * emp_g, &HFF * emp_b)
        Case &H21
            Hex2Color = RGB(&HF * emp_r, &HD7 * emp_g, &HFF * emp_b)
        Case &H22
            Hex2Color = RGB(&H69 * emp_r, &HA2 * emp_g, &HFF * emp_b)
        Case &H23
            Hex2Color = RGB(&HD4 * emp_r, &H80 * emp_g, &HFF * emp_b)
        Case &H24
            Hex2Color = RGB(&HFF * emp_r, &H45 * emp_g, &HF3 * emp_b)
        Case &H25
            Hex2Color = RGB(&HFF * emp_r, &H61 * emp_g, &H8B * emp_b)
        Case &H26
            Hex2Color = RGB(&HFF * emp_r, &H88 * emp_g, &H33 * emp_b)
        Case &H27
            Hex2Color = RGB(&HFF * emp_r, &H9C * emp_g, &H12 * emp_b)
        Case &H28
            Hex2Color = RGB(&HFA * emp_r, &HBC * emp_g, &H20 * emp_b)
        Case &H29
            Hex2Color = RGB(&H9F * emp_r, &HE3 * emp_g, &HE * emp_b)
        Case &H2A
            Hex2Color = RGB(&H2B * emp_r, &HF0 * emp_g, &H35 * emp_b)
        Case &H2B
            Hex2Color = RGB(&HC * emp_r, &HF0 * emp_g, &HA4 * emp_b)
        Case &H2C
            Hex2Color = RGB(&H5 * emp_r, &HFB * emp_g, &HFF * emp_b)
        Case &H2D
            Hex2Color = RGB(&H5E * emp_r, &H5E * emp_g, &H5E * emp_b)
        Case &H2E
            Hex2Color = RGB(&HD * emp_r, &HD * emp_g, &HD * emp_b)
        Case &H2F
            Hex2Color = RGB(&HD * emp_r, &HD * emp_g, &HD * emp_b)
        Case &H30
            Hex2Color = RGB(&HFF * emp_r, &HFF * emp_g, &HFF * emp_b)
        Case &H31
            Hex2Color = RGB(&HA6 * emp_r, &HFC * emp_g, &HFF * emp_b)
        Case &H32
            Hex2Color = RGB(&HB3 * emp_r, &HEC * emp_g, &HFF * emp_b)
        Case &H33
            Hex2Color = RGB(&HDA * emp_r, &HAB * emp_g, &HEB * emp_b)
        Case &H34
            Hex2Color = RGB(&HFF * emp_r, &HA8 * emp_g, &HF9 * emp_b)
        Case &H35
            Hex2Color = RGB(&HFF * emp_r, &HAB * emp_g, &HB3 * emp_b)
        Case &H36
            Hex2Color = RGB(&HFF * emp_r, &HD2 * emp_g, &HB0 * emp_b)
        Case &H37
            Hex2Color = RGB(&HFF * emp_r, &HEF * emp_g, &HA6 * emp_b)
        Case &H38
            Hex2Color = RGB(&HFF * emp_r, &HF7 * emp_g, &H9C * emp_b)
        Case &H39
            Hex2Color = RGB(&HD7 * emp_r, &HE8 * emp_g, &H95 * emp_b)
        Case &H3A
            Hex2Color = RGB(&HA6 * emp_r, &HED * emp_g, &HAF * emp_b)
        Case &H3B
            Hex2Color = RGB(&HA2 * emp_r, &HF2 * emp_g, &HDA * emp_b)
        Case &H3C
            Hex2Color = RGB(&H99 * emp_r, &HFF * emp_g, &HFC * emp_b)
        Case &H3D
            Hex2Color = RGB(&HDD * emp_r, &HDD * emp_g, &HDD * emp_b)
        Case &H3E
            Hex2Color = RGB(&H11 * emp_r, &H11 * emp_g, &H11 * emp_b)
        Case &H3F
            Hex2Color = RGB(&H11 * emp_r, &H11 * emp_g, &H11 * emp_b)
    End Select

End Function























'Option Base - Determines whether the lowest range of an array is 0 or 1
'  option base 1
'Erase - Erases all values of an array
'  erase (arrayname)
'Dim - Creates an array
'  dim arrayname(25)
'Redim - Resets the bounds of an array (has option to save values)
'  redim arrayname(28)
'UBound - Returns the upper dimension of an array
'  i = ubound (arrayname)
'LBound - Returns the lower dimension of an array
'  i = lbound (arrayname)
'Filter - Returns a subset of an array based on a filter
'  Filter (inputarray, searchstring)
'Array - Yes, there is a function called array. It returns an array that has been filled with data from a list. It allows you to put the actual data values in the code to avoid having the user input it or to avoid having to read it from a file
'  ArrayName = Array (10, 20, 30)
'Join - Concatenates strings within an array



' 0=ok
' 1=unknown
' 2=fail gfx conv
' 3=bad rom type

Public Function renderUniformModel(ByVal dest As PictureBox, ByVal romType As Integer, _
                                    ByRef team As clsTeam, ByVal palentry As Integer) As Integer

    Dim tiles() As Byte, result As Boolean
    
    Dim i As Integer, destx As Integer, desty As Integer
    Dim row As Integer, col As Integer
    Dim palstart As Integer, pixelval As Integer, pixelcolor As Long
    Dim nespal(3) As Long
    
    ReDim tiles(0)
    renderUniformModel = 1
    dest.Visible = False
    
    If (romType = ROM_NES) Then
    
        palstart = (palentry And 3)
        If (convertFrom2BPP(rom, &H40710, 12, tiles())) Then
            nespal(0) = Hex2Color(&H11, False, 0)
            nespal(1) = Hex2Color(&HF, False, 0)
            Select Case (palstart)
                Case 0: nespal(2) = Hex2Color(&H27, False, 0)
                        nespal(3) = Hex2Color(team.getUniformColor_NES_0(), False, 0)
                Case 1: nespal(2) = Hex2Color(&H17, False, 0)
                        nespal(3) = Hex2Color(team.getUniformColor_NES_0(), False, 0)
                Case 2: nespal(2) = Hex2Color(&H27, False, 0)
                        nespal(3) = Hex2Color(team.getUniformColor_NES_1(), False, 0)
                Case 3: nespal(2) = Hex2Color(&H17, False, 0)
                        nespal(3) = Hex2Color(team.getUniformColor_NES_1(), False, 0)
            End Select
            For i = 0 To 11
                Select Case (i)
                    Case 0: destx = 8: desty = 0
                    Case 1: destx = 8: desty = 8
                    Case 2: destx = 0: desty = 0
                    Case 3: destx = 0: desty = 8
                    Case 4: destx = 8: desty = 16
                    Case 5: destx = 8: desty = 24
                    Case 6: destx = 0: desty = 16
                    Case 7: destx = 0: desty = 24
                    Case 8: destx = 8: desty = 32
                    Case 9: destx = 8: desty = 40
                    Case 10: destx = 0: desty = 32
                    Case 11: destx = 0: desty = 40
                End Select

                For row = 0 To 7
                    For col = 0 To 7
                        pixelval = tiles((i * 64) + (row * 8) + col)
                        SetPixel dest.hdc, (destx + col), (desty + row), nespal(pixelval And 3)
                    Next col
                Next row
            Next i
        
        
        Else
            renderUniformModel = 2
            Exit Function
        End If

    
    ElseIf (romType = ROM_SNES) Then
    
        palstart = (palentry * 16)
        If (convertFrom4BPP(rom, &H88180, 4, tiles())) Then
            
            For i = 0 To 3
                Select Case (i)
                    Case 0: destx = 0: desty = 0
                    Case 1: destx = 8: desty = 0
                    Case 2: destx = 0: desty = 16
                    Case 3: destx = 8: desty = 16
                End Select
                For row = 0 To 7
                    For col = 0 To 7
                        pixelval = tiles((i * 64) + (row * 8) + col)
                        If (pixelval = 0) Then
                            pixelcolor = vbBlack
                        Else
                            pixelcolor = team.getUniformColor_RGB(palstart + (pixelval And &HF))
                        End If
                        SetPixel dest.hdc, (destx + col), (desty + row), pixelcolor
                    Next col
                Next row
            Next i
            
            If (convertFrom4BPP(rom, &H88380, 6, tiles())) Then
                For i = 0 To 5
                    Select Case (i)
                        Case 0: destx = 0: desty = 8
                        Case 1: destx = 8: desty = 8
                        Case 2: destx = 0: desty = 24
                        Case 3: destx = 8: desty = 24
                        Case 4: destx = 0: desty = 32
                        Case 5: destx = 8: desty = 32
                    End Select
                    For row = 0 To 7
                        For col = 0 To 7
                            pixelval = tiles((i * 64) + (row * 8) + col)
                            If (pixelval = 0) Then
                                pixelcolor = vbBlack
                            Else
                                pixelcolor = team.getUniformColor_RGB(palstart + (pixelval And &HF))
                            End If
                            SetPixel dest.hdc, (destx + col), (desty + row), pixelcolor
                        Next col
                    Next row
                Next i
    
                If (convertFrom4BPP(rom, &H88600, 2, tiles())) Then
                    For i = 0 To 1
                        Select Case (i)
                            Case 0: destx = 0: desty = 40
                            Case 1: destx = 8: desty = 40
                        End Select
                        For row = 0 To 7
                            For col = 0 To 7
                                pixelval = tiles((i * 64) + (row * 8) + col)
                                If (pixelval = 0) Then
                                    pixelcolor = vbBlack
                                Else
                                    pixelcolor = team.getUniformColor_RGB(palstart + (pixelval And &HF))
                                End If
                                SetPixel dest.hdc, (destx + col), (desty + row), pixelcolor
                            Next col
                        Next row
                    Next i
                Else
                    renderUniformModel = 2
                    Exit Function
                End If
                
            Else
                renderUniformModel = 2
                Exit Function
            End If
        Else
            renderUniformModel = 2
            Exit Function
        End If
    End If
    StretchBlt dest.hdc, 0, 0, 32, 96, dest.hdc, 0, 0, 16, 48, SRCCOPY
    dest.Visible = True
    renderUniformModel = 0

End Function
                            



